(* #use "topfind";; *)
(* #require "sasacore";; *)

open Sasacore

(* identifiers we expect each specific Lustre algorithm to define, once *)
let state_type = "state"
let action_type = "action"
let action_number = "actions_number"
let action_of_int = "action_of_int"

(* global setting, parsed from command-line options *)
let clock = ref false


let algo_name (node : Topology.node) = Filename.remove_extension node.file
(* prints includes, graph constants and helper functions *)
let output_prelude lustre_topology lustre_const (graph : Topology.t) =
  Printf.fprintf lustre_topology "-- automatically generated by salut \n";
  Printf.fprintf lustre_const "-- automatically generated by salut \n";
  
  (* NOTE: b2s, array_to_string and matrix_to_string copied from genOracle.ml *)
 let b2s b = if b then "t" else "f" in
 
 let (array_to_string : bool array -> string) =
  fun a ->
   let l = Array.fold_right (fun b acc -> (b2s b)::acc) a [] in
   "["^(String.concat "," l)^"]"
 in
 let (matrix_to_string : bool array array -> string) =
   fun m ->
     let l = Array.fold_right (fun a acc -> (array_to_string a)::acc) m [] in
     "[\n\t"^(String.concat ",\n\t" l)^"]"
 in
 
 (* include Lustre algos *)
 graph.nodes
 |> List.map algo_name
 |> List.sort_uniq String.compare
 |> List.iter (Printf.fprintf lustre_topology "include \"%s.lus\"\n");
 
 (* define graph constants *)
 output_string lustre_topology "\n";
 Printf.fprintf lustre_const "const card = %d;\n" (List.length graph.nodes);
 Printf.fprintf lustre_const "const links_number = %d;\n" (Topology.get_nb_link graph);
 let dmin, dmax = Topology.get_degree graph in
 Printf.fprintf lustre_const "const max_degree = %d;\n" dmax;
 Printf.fprintf lustre_const "const min_degree = %d;\n" dmin;
 Printf.fprintf lustre_const "const mean_degree = %f;\n" (Topology.get_mean_degree graph);
 Printf.fprintf lustre_const "const diameter = %d;\n" (Diameter.get graph);
 Printf.fprintf lustre_const "const is_directed = %b;\n" graph.directed;
 Printf.fprintf lustre_const "const is_cyclic = %b;\n" (Topology.is_cyclic graph);
 Printf.fprintf lustre_const "const is_connected = %b;\n" (Topology.is_connected graph);
 let neigh_list = graph.nodes |> List.map (fun (n:Topology.node) -> graph.pred n.id) in
 let nb_neigh_list = neigh_list |> List.map (fun n -> string_of_int (List.length n)) in
 Printf.fprintf lustre_const "const nb_neighbors = %s;\n" ("["^(String.concat "," (nb_neigh_list))^"]");
 
 (* dot attributes *)
 let already_defined = function
  | "card"
  | "links_number"
  | "max_degree"
  | "min_degree"
  | "mean_degree"
  | "diameter"
  | "is_directed"
  | "is_cyclic"
  | "is_connected"
  | "f"
  | "t"
  | "adjacency"  -> true
  | _ -> false in
 List.iter
  (fun (name, value) ->
   if not (already_defined name)
   then Printf.fprintf lustre_const "const %s = %s;\n" name value)
  graph.attributes;
 
 (* adjacency matrix *)
 output_string lustre_const "const t = true;\n";
 output_string lustre_const "const f = false;\n";
 Printf.fprintf lustre_const
  "const adjacency = %s;\n"
  (graph |> Topology.to_adjacency |> matrix_to_string);

 (* State.lus File *)
 output_string lustre_topology"
include \"state.lus\"";
  
 (* Neighbors type *)
 output_string lustre_topology "
type neigh = struct { state:state; reply:int; weight:int };\n";
 
 (* helper functions *)
 output_string lustre_topology "
function dot2lus_first_set<<const N:int>>(s : bool^N) returns (x : int);
var
  found : int;
let
  found =
    with (N = 1) then (if s[0] then 0 else -1)
    else dot2lus_first_set<<N-1>>(s[1 .. N-1]);
  x =
    if s[0] then 0
    else if found < 0 then -1
         else found + 1;
tel;\n";
 Printf.fprintf lustre_topology "
function dot2lus_action_of_activation(activation : bool^%s) returns (action : %s);
let
  action = %s(dot2lus_first_set<<%s>>(activation));
tel;
" action_number action_type action_of_int action_number


(* prints the actual Lustre node that implements the input topology *)
let output_topology lustre_topology (graph : Topology.t) name =
 let make_index (graph : Topology.t) : (Topology.node_id -> int) =
  let index_map = Hashtbl.create (List.length graph.nodes) in
  graph.nodes
  |> List.map (fun (n : Topology.node) -> n.id)
  |> List.iteri (fun index node_id -> Hashtbl.add index_map node_id index);
  Hashtbl.find index_map (* returns the partially applied find *) in
 
 let sprint_neighbor_list neighbor_ids list : string =
  match neighbor_ids with
    | [] -> "[]"
    | (n,r,w) :: ns ->
   let prefix, sufix = Printf.sprintf "[ neigh { state = %s[%d]; reply = %d ; weight = %d }" list n r w, " ]" in
   let concat acc (n,r,w) = acc ^ (Printf.sprintf ", neigh { state = %s[%d]; reply = %d ; weight = %d }" list n r w) in
   (List.fold_left concat prefix ns) ^ sufix in

 Printf.fprintf lustre_topology
  "\nnode %s(p : bool^%s^card; initials : %s^card)\n"
  name action_number state_type;
 
 Printf.fprintf lustre_topology
  "returns (p_c : %s^card; Enab_p : bool^%s^card; round:bool; round_nb:int);\n"
  state_type action_number;
 
 output_string lustre_topology "var";
 
 Printf.fprintf lustre_topology "\tprev_p_c : %s^card;\n\n" state_type;
 graph.nodes
 |> List.iteri (fun i _ -> Printf.fprintf lustre_topology "\tsel_%d : bool;\n" i);
 
 output_string lustre_topology "let
    round = round<<actions_number,card>>(Enab_p, p);
    round_nb = round_count(round, legitimate<<actions_number,card>>(Enab_p,p_c));
";
 
 output_string lustre_topology "\tprev_p_c = initials -> pre(p_c);\n\n";
 graph.nodes
 |> List.iteri (fun i _ ->
  Printf.fprintf lustre_topology
   "\tsel_%d = false -> boolred<<1,%s,%s>>(p[%d]);\n"
   i action_number action_number i);
 
 let index_of_id = make_index graph in
 graph.nodes
 (*liste des noms des noeuds *)
 |> List.iteri (fun i n ->
  let algo = algo_name n in
  let neighbors = graph.pred n.id |> List.map (fun (w,id) -> (index_of_id id, (Topology.reply graph n.id id),w)) in
  let deg = List.length neighbors in
  let nl = sprint_neighbor_list neighbors "p_c" in
  let pnl = sprint_neighbor_list neighbors "prev_p_c" in
  Printf.fprintf lustre_topology
   "\n\tp_c[%d] =\n\t\tif not sel_%d then prev_p_c[%d]\n\t\telse "
   i i i;
  if !clock then Printf.fprintf lustre_topology
   "current(%s_step<<%d>>(\n\t\t\tprev_p_c[%d], \n\t\t\t%s, \n\t\t\tdot2lus_action_of_activation(p[%d])\n\t\t) when sel_%d);\n"
   algo deg i pnl i i
  else Printf.fprintf lustre_topology
   "%s_step<<%d>>(\n\t\t\tprev_p_c[%d], \n\t\t\t%s, \n\t\t\tdot2lus_action_of_activation(p[%d])\n\t\t);\n"
   algo deg i pnl i;
  Printf.fprintf lustre_topology
   "\tEnab_p[%d] = %s_enable<<%d>>(p_c[%d], %s);\n"
   i algo deg i nl);
 
 Printf.fprintf  lustre_topology "tel;\nnode topology = %s;\n%!" name

(*Generate a new state.lus file*)
let state lustre_state =
   output_string lustre_state "-- automatically generated by salut \n
type state = \"define_me\";

type action = \"define_me\";
const actions_number = \"define_me\";

function action_of_int(i : int) returns (a : action);
let
  a = \"define_me\";
tel;

function legitimate<<const actions_number:int; const card:int>>(enables : bool^actions_number^card; config: state^card)
returns (res : bool);
let
  res=silent<<actions_number,card>>(enables);
tel;
"

let dot2lus dotfile lusfile =
  let graph = Topology.read dotfile in
  let name = dotfile |> Filename.basename |> Filename.chop_extension in
  let lustre_topology, lustre_const =
    match lusfile with
    | None -> open_out (name ^ ".lus"), open_out (name ^ "_const.lus")
    | Some lus -> open_out lus,
                  open_out ((Filename.remove_extension lus) ^ "_const"^(Filename.extension lus))
  in
  output_prelude lustre_topology lustre_const graph;
  output_topology lustre_topology graph name;
  close_out lustre_topology;
  close_out lustre_const;
  if Sys.file_exists "state.lus" then (
    Printf.eprintf " [salut] Warning: %s already exist.\n%!" "state.lus")
  else (
    let lustre_state = open_out ("state.lus")
    in
    state lustre_state;
    close_out lustre_state
  )


let _ =
  let usage = "salut <dotfile> [-o lusfile] [--clock] [--help]" in
  let dotfile = ref "" in
  let lusfile = ref "" in
  
  let anon_parse arg =
    match !dotfile with
    | "" -> dotfile := arg
    | _ -> raise (Arg.Bad "multiple input topologies not supported")
  in
  let speclist =
    [ ("-o", Arg.Set_string lusfile, "Set output file (default is inferred from input file)");
      ("--clock", Arg.Set clock, "Generate clocked code (default is unclocked)") ]
  in
  Arg.parse speclist anon_parse usage;
  match !dotfile with
  | "" -> print_string (Arg.usage_string speclist usage); exit 1
  | dot ->
    match !lusfile with
    | "" -> dot2lus dot None
    | lus -> dot2lus dot (Some lus)
