#
# To use this file, define the following variables and `include` it:
# TOPOLOGY     # graph file name, without the .dot extension
# DOT2LUSFLAGS # optional: dot2lus flags (eg. --clock)
# SASA_ALGOS   # .ml files used in sasa/lurette tests
# SASAFLAGS    # optional: flags passed in to sasa (eg. to define the daemon)
#

.PHONY = build all clean test verify

# extra ../ because this is called from within each test folder
D2L := salut

# default rule
build: $(TOPOLOGY).lus

all: $(TOPOLOGY).lus $(TOPOLOGY).cmxs $(TOPOLOGY)_verify.lv4

genclean:
	rm -f *.cmxs *.cmx *.cmi *.o
	rm -f $(TOPOLOGY).ml 
	rm -f *.rif *.seed
	rm -f $(TOPOLOGY).lus
	rm -f *_verify.lv4
	rm -f *.c *.h *_const.lus *.cmt *.exec *.sh rdbg-session* *.log *.pdf

# generating lv6 files of out dot files
%_const.lus: %.dot
	$(D2L) $(DOT2LUSFLAGS) $<
%.lus: %_const.lus

LIB=-package algo
OCAMLOPT=ocamlfind ocamlopt -bin-annot
LIB_LUS=$(HOME)/sasa/test/lustre/round.lus $(HOME)/sasa/salut/lib/sas.lus

# $(HOME)/sasa/test/lustre/oracle_utils.lus 


##########################################################################################"
# Simulations

%.simu:
	make $*.dot $*_const.lus
	luciole-rif lv6 $(LIB_LUS) $*.lus $*_const.lus -n $* -exec 

%.rdbg:
	make $*.dot $*_const.lus
	rdbgui4sasa -sut-nd "lv6 $(LIB_LUS) $*_const.lus $*.lus  -n $* -exec"

##########################################################################################"
# Use lurette to compare the lustre and the ocaml version of the algorithm
# nb : <dirname>/<dirname>_oracle.lus should be written

%.lurette:
	make $*.dot $*_const.lus $*.cmxs $*_oracle.lus
#	make $*.dot $*_const.lus $*.cmxs
	lurette -sut "sasa -seed 172317971 $*.dot $(SASAFLAGS)" -oracle "lv6  $(LIB_LUS) $*.lus $*_oracle.lus -n oracle -2c-exec"

%.lurette_no_seed:
	make $*.dot $*_const.lus $*.cmxs $*_oracle.lus
#	make $*.dot $*_const.lus $*.cmxs
	lurette -sut "sasa $*.dot $(SASAFLAGS)" -oracle "lv6  $(LIB_LUS) $*.lus $*_oracle.lus -n oracle -2c-exec"

%.cmxs: %.ml
	$(OCAMLOPT) $(LIB) -shared state.ml $(shell sasa -algo $*.dot) config.ml $< -o $@

# XXX le faire via salut
%_oracle.lus: %.cmxs
	sasa -glos $*.dot || echo ""

%.ml: %.dot
	sasa -reg $<


##########################################################################################"
# model-checking with kind2

ifndef $(prop)
prop = verify
endif
ifndef $(color)
color = true
endif

%.expand.kind2: %_verify.lv4
	time kind2 --color $(color) --modular true --compositional true $<

%.kind2: %.dot
	make $*_verify.noexpand.lv4
	time kind2 --color $(color) --modular true --compositional true $*_verify.noexpand.lv4

%.kind2-test:
	make color=false $*.kind2  | (grep "^ok: valid" > $*-kind2.res || echo "") ; [  -s $*-kind2.res ]

# XXX: this method of adding kind2 annotations is brittle
%_verify.lv4: %.lus %_const.lus  verify.lus
	lv6 -eei $(LIB_LUS) $*.lus $^ $*_const.lus -n $(prop) --lustre-v4 -en -o $@
	sed -i -e "s/tel/--%MAIN ;\n--%PROPERTY ok;\ntel/" $@

# Do not expand nodes 
%_verify.noexpand.lv4:  verify.lus
	make $*_const.lus
	lv6 -eei $(LIB_LUS) $*.lus $^ $*_const.lus -n $(prop) -rnc --lustre-v4 -o $@
	cat $@ | tr '\n' '@' |sed "s/tel@/tel/g" | sed "s/@/\\n/g" | sed "s/tel-- end of node verify/--%MAIN ;\n--%PROPERTY ok;\ntel\n/"  > $@.tmp
	mv $@.tmp $@

##########################################################################################"
# model-checking with lesar

# to use lesar, perform bit-blasting via ec2ec -usrint
%_verify.ec: %.dot verify.lus
	make $*_const.lus
	lv6  $(LIB_LUS) $*.lus $*_const.lus verify.lus -n $(prop) -ec -o $@.tmp
	echo "include \"../../lib/bit-blast/signed_binary6.lus\"" > $@.tmp2
	echo "include \"../../lib/bit-blast/binary.lus\"" >> $@.tmp2
	ec2ec -usrint $@.tmp >> $@.tmp2
	cat $@.tmp2 | tr '\n' '@' |sed "s/tel@/tel/g" | sed "s/@/\\n/g" | sed "s/tel-- end of node verify/--%MAIN ;\n--%PROPERTY ok;\ntel\n/"  > $@.tmp
	mv $@.tmp $@ ; rm  $@.tmp2


%.lesar:%_verify.ec
	time lesar $< $(prop) -forward -states 10000000 

##########################################################################################"
# Compare sasa and salut simulation perf
# the sasa version is called twice: once for generating the daemon, and once
# to perform a fair measurement wrt salut
# XXX lurette kills the process before time can write its results!
# Anyway, it's not the good way to perform time measurements. It is
# better to use internal daemon 

%.comperf.old: %.dot %.cmxs %_oracle.lus %.lus
	lv6 -en -2c $*.lus $*_oracle.lus -n oracle && \
	sh oracle.sh && \
	/usr/bin/time -o sasa2.time lurette -norif -env "sasa -seed 42 -custd $*.dot $(SASAFLAGS) " \
		-sut "sasa -seed 42 $*.dot $(SASAFLAGS)" &&\
	/usr/bin/time -o salut.time lurette -norif -env "sasa -seed 42 -custd $*.dot $(SASAFLAGS) " \
		-oracle "./oracle.exec" \
		-sut "sasa -seed 42 $*.dot $(SASAFLAGS)" && \
	cat sasa.time sasa2.time salut.time 

### to compute the array in Section VIII-B
synchronous_diring500.exec: diring500.dot
	make diring500.lus
	cat synchronous_diring500.lus >> diring500.lus
	lv6 -2c diring500.lus -n synchronous_diring500
	sh synchronous_diring500.sh

run_synchronous_diring500.salut: synchronous_diring500.exec
	for ((a=1; a <= 500 ; a++)) do echo f; done | time ./synchronous_diring500.exec > /tmp/x.rif

run_synchronous_diring500.sasa: diring500.dot diring500.cmxs
	time sasa diring500.dot -sd -l 500 > /tmp/x.rif

run_synchronous_er500.exec: er500.dot
	make er500.lus
	cat synchronous_er500.lus >> er500.lus
	lv6 -2c er500.lus -n synchronous_er500
	sh synchronous_er500.sh

grid500.dot:
	gg grid -w 20 -he 25 -o grid500.dot
	gg-deco "0-:p.ml" grid500.dot -o  grid500.dot

run_synchronous_er500.salut: synchronous_er500.exec
	for ((a=1; a <= 500 ; a++)) do echo f; done | time ./synchronous_er500.exec > /tmp/x.rif
run_synchronous_grid500.salut: synchronous_grid500.exec
	for ((a=1; a <= 500 ; a++)) do echo f; done | time ./synchronous_grid500.exec > /tmp/x.rif

run_synchronous_er500.sasa: er500.dot er500.cmxs
	time sasa er500.dot -sd -l 500 > /tmp/x.rif

